#ifndef BITBULLY_VERSION_H
#define BITBULLY_VERSION_H

#define PROJECT_VERSION "0.0.75"
#define PROJECT_MAJOR_VERSION "0"
#define PROJECT_MINOR_VERSION "0"
#define PROJECT_PATCH_VERSION "75"

#endif  // BITBULLY_VERSION_H
